# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#



from duHast.Utilities.date_stamps import FILE_DATE_STAMP_YYYY_MM_DD, get_date_stamp
from duHast.Revit.Family.Data.Objects.family_directive_copy import FamilyDirectiveCopy
from duHast.Revit.Family.Data.Objects.family_directive_swap_instances_of_type import FamilyDirectiveSwap

# the parameter containing the grouping code in each family
GROUPING_CODE_PARAMETER_NAME = "HSL_AHFG_CODE"

# file name of swap directives
SWAP_DIRECTIVE_FILE_NAME = "{} {}{}".format(
    FamilyDirectiveSwap.SWAP_DIRECTIVE_FILE_NAME_PREFIX, 
    get_date_stamp(FILE_DATE_STAMP_YYYY_MM_DD),
    FamilyDirectiveSwap.SWAP_DIRECTIVE_FILE_EXTENSION
)

# prefix for the file name of maintain family types by family directives
MAINTAIN_TYPES_BY_FAMILY_FILE_NAME_PREFIX = "maintain_types_by_family"
# file name of maintain family types by family directives
MAINTAIN_TYPES_BY_FAMILY_FILE_NAME = "{} {}.csv".format(MAINTAIN_TYPES_BY_FAMILY_FILE_NAME_PREFIX, get_date_stamp(FILE_DATE_STAMP_YYYY_MM_DD))

# file name for copy directive
COPY_DIRECTIVE_FILE_NAME = "{} {}{}".format(
    FamilyDirectiveCopy.COPY_DIRECTIVE_FILE_NAME_PREFIX, 
    get_date_stamp(FILE_DATE_STAMP_YYYY_MM_DD), 
    FamilyDirectiveCopy.COPY_DIRECTIVE_FILE_EXTENSION
)

# file name for duplicate copy directives
DUPLICATE_COPY_DIRECTIVE_FILE_NAME = "{} {}{}".format(
    "duplicate__"+FamilyDirectiveCopy.COPY_DIRECTIVE_FILE_NAME_PREFIX, 
    get_date_stamp(FILE_DATE_STAMP_YYYY_MM_DD), 
    FamilyDirectiveCopy.COPY_DIRECTIVE_FILE_EXTENSION
)

# file name for families with missing group codes
FAMILIES_WITH_MISSING_GROUP_CODES_FILE_NAME = "missing_group_codes {}.csv".format(get_date_stamp(FILE_DATE_STAMP_YYYY_MM_DD))

# file name prefix depending on the family category
CATEGORY_FILE_NAME_PREFIX_MAPPER = {
    "Audio Visual Devices": "AVD",
    "Casework": "CSW",
    "Data Devices": "DAT",
    "Electrical Equipment": "ELE",
    "Electrical Fixtures": "ELF",
    "Fire Alarm Devices": "FIR",
    "Furniture": "FRN",
    "Furniture Systems": "FRS",
    "Generic Models": "GEN",
    "Lighting Devices": "LGD",
    "Lighting Fixtures": "LGF",
    "Mechanical Equipment": "MEC",
    "Medical Equipment": "MEF",
    "Nurse Call Devices": "NRS",
    "Plumbing Fixtures": "PLM",
    "railings": "RLG",
    "Security Devices": "SEC",
    "Specialty Equipment": "SPC",
    "Sprinklers": "SPR",
    "Telephone Devices": "TEL",
    "Windows": "WDW",
}


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Data storage reader class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


# import clr
# clr.AddReference("System.Core")
# from System import Linq
# clr.ImportExtensions(Linq)

from duHast.Data.Objects.Collectors import data_ceiling as dc
from duHast.Data.Objects.Collectors import data_room as dr
from duHast.Data.Utils.data_to_file import CONSTANT_DATA_FIELDS
from duHast.Utilities.files_json import read_json_data_from_file


class ReadDataFromFile:
    """
    Class to read data from a json formatted file and store it in data objects.

    The class supports the following data types:

    - :class: `.DataRoom`
    - :class: `.DataCeiling`

    """

    def __init__(self, file_path):
        """
        Class constructor.

        :param filePath: Fully qualified file path to json formatted data file.
        :type filePath: str
        """

        self.data_file_path = file_path
        self.data_type = ""
        self.data = []
        self.data_by_type = {}
        self.debug_messages = []

    # list of data types supported by this class
    SUPPORTED_DATA_TYPES = {
        dr.DataRoom.data_type: dr.DataRoom,
        dc.DataCeiling.data_type: dc.DataCeiling,
    }

    def _load_data_type(self, json_object, data_type_name):
        """
        Load data type from json object.

        :param json_object: Dictionary containing data type.
        :type json_object: dict
        :param data_type: The data type name.
        :type data_type: str
        """

        if data_type_name in json_object:
            data_type_class = self.SUPPORTED_DATA_TYPES.get(data_type_name)
            # check if supported data type
            if data_type_class is None:
                self.debug_messages.append(
                    "Data type not supported: {}".format(data_type_name)
                )
                raise ValueError("Data type not supported: {}".format(data_type_name))

            # load data
            for d in json_object[data_type_name]:
                # self.debug_messages.append("Loading data type: {}".format(data_type_name))
                p = data_type_class(d)
                # append to data by type dictionary
                if data_type_name in self.data_by_type:
                    self.data_by_type[data_type_name].append(p)
                else:
                    self.data_by_type[data_type_name] = [p]
                # append to overall data list
                self.data.append(p)

        else:
            pass
            # not all data types are always present in the json object!

    def add_debug_message_from_load(self, json):
        """
        Returns a list of debug messages from the load process.

        :return: A list of debug messages.
        :rtype: list
        """
        # debug message values
        # get fixed values
        file_name = "not set"
        date_processed = "not set"
        if CONSTANT_DATA_FIELDS["file name"] in json:
            file_name = json[CONSTANT_DATA_FIELDS["file name"]]
        if CONSTANT_DATA_FIELDS["date processed"] in json:
            date_processed = json[CONSTANT_DATA_FIELDS["date processed"]]
        # get other keys
        other_key_values = []
        other_keys = list(json.keys())
        for data_type_name in self.SUPPORTED_DATA_TYPES:
            if data_type_name in other_keys:
                other_key_values.append(
                    "{} :{} values".format(data_type_name, len(json[data_type_name]))
                )

        # build debug message
        self.debug_messages.append(
            "...Loading data from dictionary.  \n......File name: {}\n......Date processed: {}\n......{}".format(
                file_name, date_processed, "\n......".join(other_key_values)
            )
        )

    def load_data(self):
        """
        Load json formatted rows into data objects and stores them in this class.

        In the moment the following data objects are supported:

        - :class: `.DataRoom`
        - :class: `.DataCeiling`

        """

        # load json from file
        read_result = read_json_data_from_file(self.data_file_path)
        if read_result.status is False:
            self.debug_messages.append(
                "Failed to read data from file: {}".format(read_result.message)
            )
            raise ValueError(read_result.message)
        data_json = read_result.result[0]
        
        # this data can com in two different formats:
        # 1. a list of dictionaries as per item 2. below
        # 2. a dictionary with 4 keys: 'date processed', 'room', 'ceiling', 'file name'.'room'  and 'ceiling' are lists of dictionaries as  values
        # the first format is the result of multiple exported files , i.e. one per level being merged into one report file ie. for the entire building

        if len(data_json) > 0:
            # check type of data
            if isinstance(data_json, list):
                self.debug_messages.append(
                    "Data format: list of dictionaries with: {} list entries".format(
                        len(data_json)
                    )
                )
                # we have a list of dictionaries
                for entry in data_json:
                    # add debug message for this json entry
                    self.add_debug_message_from_load(entry)

                    # load each  data type
                    for data_type_name in self.SUPPORTED_DATA_TYPES:
                        self._load_data_type(
                            json_object=entry,
                            data_type_name=data_type_name,
                        )

            elif isinstance(data_json, dict):
                self.debug_messages.append(
                    "Data format: Single dictionary with: {} keys".format(
                        len(data_json)
                    )
                )
                # add debug message for this json entry
                self.add_debug_message_from_load(data_json)

                # we have a dictionary
                # load each  data type
                for data_type_name in self.SUPPORTED_DATA_TYPES:
                    self._load_data_type(
                        json_object=data_json,
                        data_type_name=data_type_name,
                    )
            else:
                raise TypeError("Data format not supported: {}".format(type(data_json)))

        else:
            self.debug_messages.append(
                "No data found in file: {}".format(self.data_file_path)
            )

        # final load stats
        self.debug_messages.append(
            "Data loaded: {} data objects".format(len(self.data))
        )
        for data_type_name in self.SUPPORTED_DATA_TYPES:
            if data_type_name in self.data_by_type:
                self.debug_messages.append(
                    "...Data type {} has: {} entries".format(
                        data_type_name, len(self.data_by_type[data_type_name])
                    )
                )
            else:
                self.debug_messages.append(
                    "...Data type {} has: 0 entries".format(data_type_name)
                )

    def get_data_by_level(self, level_name):
        """
        Returns all data objects where level name equals past in value.

        :param level_name: The building level name.
        :type level_name: str

        :return: A list of room and ceiling data objects
        :rtype: list [data objects]
        """

        return list(filter(lambda x: (x.level.name == level_name), self.data))

    def get_data_by_type(self, data_type):
        """
        Returns all data objects where type equals past in type name

        :param data_type: The data type name.
        :type data_type: str

        :return: A list of room and ceiling data objects
        :rtype: list [data objects]
        """

        return list(filter(lambda x: (x.data_type == data_type), self.data))

    def get_data_by_level_and_data_type(self, level_name, data_type):
        """
        Returns all data objects where level name and data type equal past in values.

        :param level_name: The building level name.
        :type level_name: str
        :param data_type: A string describing the data type\
            refer to property .dataType on data object class
        :type data_type: str

        :return: A list of data objects
        :rtype: list [data objects]
        """

        return list(
            filter(
                lambda x: (x.level.name == level_name and x.data_type == data_type),
                self.data,
            )
        )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging unused mep system types.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common import purge_utils as rPurgeUtils
from duHast.Revit.MEP_Systems.cable_trays import (
    get_all_cable_tray_type_ids_in_model_by_category,
    get_all_cable_tray_types_by_category,
)
from duHast.Revit.MEP_Systems.conduits import (
    get_all_conduit_type_ids_in_model_by_category,
    get_all_conduit_types_by_category,
)
from duHast.Revit.MEP_Systems.ducts import (
    get_all_duct_type_ids_in_model_by_category,
    get_all_duct_types_by_category,
)
from duHast.Revit.MEP_Systems.flex_ducts import (
    get_all_flex_duct_type_ids_in_model_by_category,
    get_all_flex_duct_types_by_category,
)
from duHast.Revit.MEP_Systems.pipes import (
    get_all_pipe_type_ids_in_model_by_category,
    get_all_pipe_types_by_category,
)
from duHast.Revit.MEP_Systems.Utility.RevitMEPSystemNames import (
    BUILTIN_CABLE_TRAY_TYPE_FAMILY_NAMES,
    BUILTIN_CONDUIT_TYPE_FAMILY_NAMES,
    BUILTIN_DUCT_TYPE_FAMILY_NAMES,
    BUILTIN_FLEX_DUCT_TYPE_FAMILY_NAMES,
    BUILTIN_PIPE_TYPE_FAMILY_NAMES,
)
from duHast.Revit.MEP_Systems.Utility.RevitMEPTypeSorting import (
    sort_types_by_family_name,
)


def get_used_duct_type_ids(doc):
    """
    Gets all used duct type ids available in model.
    Unused: at least instance of each of these types is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing duct types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_duct_type_ids_in_model_by_category, 1
    )
    return ids


def get_used_flex_duct_type_ids(doc):
    """
    Gets all used flex duct type ids available in model.
    Unused: at least instance of each of these types is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing flex duct types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_flex_duct_type_ids_in_model_by_category, 1
    )
    return ids


def get_used_conduit_type_ids(doc):
    """
    Gets all used conduit type ids available in model.
    Unused: at least instance of each of these types is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing conduit types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_conduit_type_ids_in_model_by_category, 1
    )
    return ids


def get_used_cable_tray_type_ids(doc):
    """
    Gets all used cable tray type ids available in model.
    Unused: at least instance of each of these types is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing cable tray types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_cable_tray_type_ids_in_model_by_category, 1
    )
    return ids


def get_used_pipe_type_ids(doc):
    """
    Gets all used pipe type ids available in model.
    Unused: at least instance of each of these types is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing pipe types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_pipe_type_ids_in_model_by_category, 1
    )
    return ids


def get_unused_duct_type_ids(doc):
    """
    Gets all unused duct type ids available in model.
    Unused: not one instance of these types is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing duct types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_duct_type_ids_in_model_by_category, 0
    )
    return ids


def get_unused_flex_duct_type_ids(doc):
    """
    Gets all unused flex duct type ids available in model.
    Unused: not one instance of these types is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing flex duct types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_flex_duct_type_ids_in_model_by_category, 0
    )
    return ids


def get_unused_conduit_type_ids(doc):
    """
    Gets all unused conduit type ids available in model.
    Unused: not one instance of these types is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing conduit types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_conduit_type_ids_in_model_by_category, 0
    )
    return ids


def get_unused_cable_tray_type_ids(doc):
    """
    Gets all unused cable tray type ids available in model.
    Unused: not one instance of these types is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing cable tray types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_cable_tray_type_ids_in_model_by_category, 0
    )
    return ids


def get_unused_pipe_type_ids(doc):
    """
    Gets all unused pipe type ids available in model.
    Unused: not one instance of these types is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing pipe types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_pipe_type_ids_in_model_by_category, 0
    )
    return ids


def family_no_types_in_use(fam_type_ids, un_used_type_ids):
    """
    Compares two lists of element ids and returns False if any element id in first list is not in the second list.
    Returns False if any symbols (types) of a family (first list) are in use in a model (second list).
    TODO: repetitive code...Consider generic function!
    :param fam_type_ids: List of family symbols (types).
    :type fam_type_ids: List of Autodesk.Revit.DB.ElementId
    :param un_used_type_ids: List of unused family symbols (types)
    :type un_used_type_ids: List of Autodesk.Revit.DB.ElementId
    :return: True if all ids in first list are also in second list, otherwise False.
    :rtype: bool
    """

    match = True
    for fam_type_id in fam_type_ids:
        if fam_type_id not in un_used_type_ids:
            match = False
            break
    return match


def get_unused_mep_system_type_ids_to_purge(
    doc, all_type_id_getter, all_types_getter, built_in_family_type_names
):
    """
    Gets the ids of unused MEP system types. 
    In the case that no mep system instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one mep system type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param all_type_id_getter: Function getting all available system type ids as a list.
    :type all_type_id_getter: func(doc) -> List Autodesk.Revit.ED.ElementId
    :param all_types_getter: Function getting all available system types as a collector.
    :type all_types_getter: func(doc) -> Autodesk.Revit.DB.FilteredElementCollector
    :param built_in_family_type_names: List containing all available major type names.
    :type built_in_family_type_names: List str
    :return: A list of ids representing mep system types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(doc, all_type_id_getter, 0)
    # make sure there is at least on Stair type per system family left in model
    types = sort_types_by_family_name(doc, all_types_getter)
    for key, value in types.items():
        if key in built_in_family_type_names:
            if family_no_types_in_use(value, ids) == True:
                # remove one type of this system family from unused list
                ids.remove(value[0])
    return ids


def get_unused_duct_type_ids_to_purge(doc):
    """
    Gets all unused duct type ids. 
    This method can be used to safely delete unused duct types:
    In the case that no duct instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one duct type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing duct types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = get_unused_mep_system_type_ids_to_purge(
        doc,
        get_all_duct_type_ids_in_model_by_category,
        get_all_duct_types_by_category,
        BUILTIN_DUCT_TYPE_FAMILY_NAMES,
    )
    return ids


def get_unused_flex_duct_type_ids_to_purge(doc):
    """
    Gets all unused flex duct type ids. 
    This method can be used to safely delete unused flex duct types:
    In the case that no flex duct instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one flex duct type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing flex duct types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = get_unused_mep_system_type_ids_to_purge(
        doc,
        get_all_flex_duct_type_ids_in_model_by_category,
        get_all_flex_duct_types_by_category,
        BUILTIN_FLEX_DUCT_TYPE_FAMILY_NAMES,
    )
    return ids


def get_unused_conduit_type_ids_to_purge(doc):
    """
    Gets all unused conduit type ids. 
    This method can be used to safely delete unused conduit types:
    In the case that no conduit instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one conduit type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing conduit types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = get_unused_mep_system_type_ids_to_purge(
        doc,
        get_all_conduit_type_ids_in_model_by_category,
        get_all_conduit_types_by_category,
        BUILTIN_CONDUIT_TYPE_FAMILY_NAMES,
    )
    return ids


def get_unused_cable_tray_type_ids_to_purge(doc):
    """
    Gets all unused cable tray type ids. 
    This method can be used to safely delete unused cable tray types:
    In the case that no cable tray instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one cable tray type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing conduit types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = get_unused_mep_system_type_ids_to_purge(
        doc,
        get_all_cable_tray_type_ids_in_model_by_category,
        get_all_cable_tray_types_by_category,
        BUILTIN_CABLE_TRAY_TYPE_FAMILY_NAMES,
    )
    return ids


def get_unused_pipe_type_ids_to_purge(doc):
    """
    Gets all unused pipe type ids. 
    This method can be used to safely delete unused pipe types:
    In the case that no pipe instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one pipe type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing conduit types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = get_unused_mep_system_type_ids_to_purge(
        doc,
        get_all_pipe_type_ids_in_model_by_category,
        get_all_pipe_types_by_category,
        BUILTIN_PIPE_TYPE_FAMILY_NAMES,
    )
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A bounding box base class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json
from duHast.Utilities.Objects.base import Base
from duHast.Geometry.geometry_property_names import GeometryPropertyNames
from duHast.Utilities.compare import is_close

class BoundingBoxBase(Base):
    def __init__(self, j=None, **kwargs):
        """
        Base implementation of a bounding box.

        :raises TypeError: "Input must be a JSON string or a dictionary."
        :raises ValueError: "JSON must contain 'point1' and 'point2' keys."
        """
        # ini super class to allow multi inheritance in children!
        super(BoundingBoxBase, self).__init__(**kwargs)

        json_string=None
        # Check if a JSON string / dictionary is provided
        if j:
            if isinstance(j, str):
                # Parse the JSON string
                json_string = json.loads(j)
            elif isinstance(j, dict):
                # make a copy
                json_string=j.copy()
            elif not isinstance(j, dict):
                raise TypeError("Input must be a JSON string or a dictionary.")

            # Validate presence of required keys
            if (
                GeometryPropertyNames.MAX_X not in json_string
                or GeometryPropertyNames.MAX_Y not in json_string
                or GeometryPropertyNames.MIN_X not in json_string
                or GeometryPropertyNames.MIN_Y not in json_string
            ):
                raise ValueError("JSON must contain 'max_x', 'max_y', 'min_x', 'min_y' keys.")
            self._json_ini = json_string
        else:
            self._json_ini = None

        # set default values
        self._min_x = 0.0
        self._max_x = 0.0
        self._min_y = 0.0
        self._max_y = 0.0

    @property
    def json_ini(self):
        """Read-only property to access the parsed JSON data."""
        return self._json_ini

    @property
    def min_x(self):
        """Read-only property for minimum x value."""
        return self._min_x

    @property
    def max_x(self):
        """Read-only property for maximum x value."""
        return self._max_x

    @property
    def min_y(self):
        """Read-only property for minimum y value."""
        return self._min_y

    @property
    def max_y(self):
        """Read-only property for maximum y value."""
        return self._max_y

    def contains(self, point):
        raise NotImplementedError("Subclasses should implement this method")

    def width(self):
        raise NotImplementedError("Subclasses should implement this method")

    def depth(self):
        raise NotImplementedError("Subclasses should implement this method")
    
    def ratio(self):
        raise NotImplementedError("Subclasses should implement this method")
        
    def __str__(self):
        return "BoundingBoxBase({}, {}, {}, {})".format(
            self.min_x, self.min_y, self.max_x, self.max_y
        )

    def __eq__(self, other):
        if not isinstance(other, BoundingBoxBase):
            return NotImplemented
        return (is_close(self.min_x, other.min_x) and
                is_close(self.max_x, other.max_x) and
                is_close(self.min_y, other.min_y) and
                is_close(self.max_y, other.max_y))

    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __hash__(self):
        return hash((self.min_x, self.max_x, self.min_y, self.max_y))


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.result import Result

from Autodesk.Revit.DB import ElementClassFilter, FilteredElementCollector, ColorFillScheme

def get_all_colour_schemes(doc):
    """Returns a list of available colour schemes."""
    color_fill_schemes = FilteredElementCollector(doc).OfClass(ColorFillScheme)
    return color_fill_schemes


def get_colour_fill_scheme_from_area_scheme(doc, area_scheme):
    """Returns a colour fill scheme from an area scheme."""
    
    # set up an element filter to find the dependent elements
    dependent_element_filter = ElementClassFilter(ColorFillScheme)

    # get alll area fill schemes
    dependent_element_ids = area_scheme.GetDependentElements(dependent_element_filter)

    # get the area fill schemes from the ids
    colour_fill_schemes = []

    if(dependent_element_ids is None):
        colour_fill_schemes

    # loop over the dependent element ids and get the colour fill schemes
    for dependent_element_id in dependent_element_ids:
        colour_fill_scheme = doc.GetElement(dependent_element_id)
        if colour_fill_scheme is not None:
            colour_fill_schemes.append(colour_fill_scheme)
    
    return colour_fill_schemes


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A vector base class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import math

from duHast.Utilities.Objects.base import Base
from duHast.Geometry.Exceptions.incompatible_vector_dimension import (
    IncompatibleVectorDimensions,
)
from duHast.Utilities.compare import is_close

class VectorBase(Base):
    def __init__(self, *components):
        """
        A vector base class.
        """
        # ini super class to allow multi inheritance in children!
        super(VectorBase, self).__init__()
        
        # check components are either int ort floats
        for component in components:
            if not isinstance(component, (float, int)):
                raise TypeError(
                    "All components must be of type float or int, got {} instead.".format(type(component).__name__)
                )

        self.components = components

    def _check_dimension_compatibility(self, other):
        if len(self.components) != len(other.components):
            raise IncompatibleVectorDimensions(
                "Dimension mismatch: {} vs {}".format(
                    len(self.components), len(other.components)
                ),
                other,
            )

    def magnitude(self):
        return math.sqrt(sum(c**2 for c in self.components))

    def __str__(self):
        return "Vector ({})".format(", ".join(map(str, self.components)))

    def __rmul__(self, s):
        return self.__mul__(s)

    def __neg__(self):
        # Raise a NotImplementedError to force child classes to implement their own
        raise NotImplementedError("Negation is not implemented in VectorBase; child classes must provide their own implementation.")

    def __pos__(self):
        return self

    def __abs__(self):
        return self.magnitude()

    def __eq__(self, other):
        if not isinstance(other, VectorBase):
            return NotImplemented
        self._check_dimension_compatibility(other)
        return all(is_close(a, b) for a, b in zip(self.components, other.components))
    
    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __hash__(self):
        return hash(self.components)


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to writing text files. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import codecs
import csv

from duHast.Utilities.files_io import (
    is_last_char_newline,
)

from duHast.Utilities.Objects.result import Result


def write_report_data(
    file_name,
    header,
    data,
    write_type="w",
    enforce_ascii=False,
    encoding="utf-8",
    bom=None,
    quoting=csv.QUOTE_NONE,
    delimiter=",",
):
    """
    Function writing out report information as CSV file.
    :param file_name: The reports fully qualified file path.
    :type file_name: str
    :param header: list of column headers
    :type header: list of str
    :param data: list of list of strings representing row data
    :type data: [[str,str,..]]
    :param write_type: Flag indicating whether existing report file is to be overwritten 'w' or appended to 'a', defaults to 'w'
    :type write_type: str, optional
    :param enforce_ascii: Flag to enforce ASCII encoding on data. If True, data will be encoded to ASCII. Defaults to False.
    :type enforce_ascii: bool, optional
    :param encoding: Encoding used to write the file. Defaults to 'utf-8'. None no encoding applied.
    :type encoding: str, optional
    :param bom: the byte order mark, Default is None (none will be written). BOM: "utf-16" = , "utf-16-le" = ,  utf-8 =
    :type bom: str, default is NoneType
    :param quoting: Quoting style used by the csv writer. Defaults to csv.QUOTE_NONE. Options are csv.QUOTE_ALL, csv.QUOTE_MINIMAL, csv.QUOTE_NONNUMERIC, csv.QUOTE_NONE
    :type quoting: int, optional

    :return:
        Result class instance.

        - result.status (bool) True if file was written without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will be an empty list.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()

    # set a flag to check if we need to add a newline before writing
    need_newline = False
    # if in append mode, check if the last character is a newline
    if write_type == "a":
        if not is_last_char_newline(file_name):
            # if not, we need to add a newline before writing
            need_newline = True
            return_value.append_message(
                "File: {} is in append mode, but last character is not a newline.".format(
                    file_name
                )
            )

    # if no encoding is provided set the encoding to ascii (default)
    encoding_file_open = encoding if encoding is not None else "ascii"
    
    # Open the file with the codecs.open method to specify encoding
    with codecs.open(file_name, write_type, encoding=encoding_file_open) as f:
        try:
            # Write a newline character if appending to the file to make sure we are starting on a new line
            if write_type == "a" and need_newline:
                f.write("\n")

            # Write BOM manually if specified
            if bom and "w" in write_type:
                f.write(bom.decode(encoding))

            # Create the CSV writer
            # line terminator is set to '\n' to avoid double newlines on Windows
            writer = csv.writer(
                f,
                delimiter=delimiter,
                escapechar="\\",
                quoting=quoting,
                lineterminator="\n",
            )

            # internal function to encode the row using the specified encoding
            def encoded_row(row):
                # Encode each string in the row using the specified encoding
                encoded = [s.encode(encoding).decode(encoding) for s in row]
                # check if also ascii encoding is enforced
                if enforce_ascii:
                    # enforce ascii encoding by removing non-ascii characters
                    return [
                        s.encode("ascii", "ignore").decode("ascii") for s in encoded
                    ]
                else:
                    # return the encoded row
                    return (
                        encoded  # Keep the strings in their current state for writing
                    )

            # Write header
            wrote_header = False
            if header and len(header) > 0:
                # check if encoding is required
                if encoding is not None:
                    # Encode the header using the specified encoding
                    header = encoded_row(header)
                writer.writerow(header)
                return_value.append_message(
                    "Header written to file. (including newline)"
                )
                # set flag that header was written
                wrote_header = True
            else:
                return_value.append_message("No header provided, skipping writing.")

            # Write data rows
            wrote_date = False
            if data and len(data) > 0:
                for i in range(len(data)):
                    row = data[i]
                    
                    # check if encoding is required
                    if encoding is not None:
                        # Encode the row using the specified encoding
                        row = encoded_row(data[i])
                    
                    # write the row data
                    # ( do not log what got written into results message property since that is a massive performance hit in the moment )
                    writer.writerow(row)
                    
                # set flag that data was written
                wrote_date = True
            else:
                return_value.append_message("No data provided, skipping writing.")

            # Remove the newline character from the last row if any data was written
            if wrote_date or wrote_header:
                f.flush()  # Ensure all data is written to the file
                with open(file_name, "rb+") as f:
                    f.seek(-1, 2)  # Move the cursor to the last character in the file
                    if f.read(1) == b"\n":
                        f.seek(-1, 2)  # Move the cursor back by one character
                        f.truncate()  # Truncate the file at the current cursor position
                        return_value.append_message(
                            "Removed newline character from the last row."
                        )

        except Exception as e:
            return_value.update_sep(
                False,
                "File: {} failed to write data with exception: {}".format(file_name, e),
            )
        finally:
            # make sure to close the file
            if (f):
                f.close()

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Duplicate mark warnings solver class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value
from duHast.Revit.Common.parameter_set_utils import (
    set_builtin_parameter_without_transaction_wrapper_by_name,
)
from duHast.Revit.Common.Objects.FailureHandlingConfiguration import (
    FailureHandlingConfig,
)
from duHast.Revit.Common.transaction import in_transaction_with_failure_handling
from duHast.Utilities.Objects import result as res
from duHast.Revit.Warnings.warning_guids import DUPLICATE_MARK_VALUE
from duHast.Utilities.Objects import base

# import Autodesk
from Autodesk.Revit.DB import BuiltInParameter, Element, Transaction


class RevitWarningsSolverDuplicateMark(base.Base):
    def __init__(
        self,
        filter_func,
        filter_values=None,
        callback=None,
    ):
        """
        Constructor: this solver takes two arguments: a filter function and a list of values to filter by

        :param filter_func: A function to filter elements in warnings by
        :type filter_func: func(document, elementId, list of filter values)
        :param filter_values: A list of filter values, defaults to []
        :type filter_values: list, optional
        """

        # ini super class to allow multi inheritance in children!
        super(RevitWarningsSolverDuplicateMark, self).__init__()

        self.filter = filter_func
        # check if default value
        if filter_values == None:
            self.filter_values = []
        else:
            self.filter_values = filter_values

        self.filter_name = "Duplicate mark value."
        self.callback = callback

    # --------------------------- duplicate mark guid ---------------------------
    #: guid identifying this specific warning
    GUID = DUPLICATE_MARK_VALUE

    IGNORED_WARNINGS = ["Type Mark"]

    def _setup_failure_handling_config(self):
        # define failure handling for the transaction ( push through on any warnings or errors )
        failure_handling_settings = FailureHandlingConfig(
            roll_back_on_warning=False,
            print_warnings=False,
            roll_back_on_error=False,
            print_errors=False,
        )
        return failure_handling_settings

    def solve_warnings(self, doc, warnings):
        """
        Solver setting element mark to nothing, provided it passes the filter.

        :param doc: Current Revit model document.
        :type doc: Autodesk.Revit.DB.Document
        :param warnings: List of warnings to be solved.
        :type warnings: Autodesk.Revit.DB.FailureMessage

        :return:
            Result class instance.

            - .status True if all duplicate mark warnings could be solved. Otherwise False.
            - .message will contain stats in format parameter value set to ''

        :rtype: :class:`.Result`
        """

        return_value = res.Result()
        if len(warnings) > 0:

            # report progress to call back if required
            counter = 1

            for warning in warnings:

                # report progress to call back if required
                if self.callback:
                    self.callback.update(counter, len(warnings))

                # Flag to indicate if we should continue the outer loop
                # ignoring this warning
                should_continue_outer = False

                # check for any duplicate Type Mark warnings...which are not to be addressed!
                for ignore in self.IGNORED_WARNINGS:
                    if ignore in warning.GetDescriptionText():
                        element_ids = warning.GetFailingElements()
                        for el_id in element_ids:
                            element = doc.GetElement(el_id)
                            return_value.update_sep(
                                True,
                                "{} Warning of type: duplicate {}. {} will be ignored.".format(
                                    self.filter_name,
                                    Element.Name.GetValue(element),
                                    ignore,
                                ),
                            )
                        should_continue_outer = True
                        break  # Break out of the inner loop (ignore this warning)
                if should_continue_outer:
                    continue  # Continue the outer loop

                # loop over elements in warning and set mark to an empty value
                element_ids = warning.GetFailingElements()
                # set up a failure handling config ignoring all warnings and errors
                failure_handling_config = self._setup_failure_handling_config()

                for el_id in element_ids:
                    element = doc.GetElement(el_id)
                    # check whether element passes filter
                    if self.filter(doc, el_id, self.filter_values):

                        try:
                            p_value = get_built_in_parameter_value(
                                element, BuiltInParameter.ALL_MODEL_MARK
                            )
                            if p_value != None:
                                # set up an action to run in a transaction with custom failure handling
                                def action():
                                    action_return_value = res.Result()
                                    try:
                                        action_return_value = set_builtin_parameter_without_transaction_wrapper_by_name(
                                            element=element,
                                            parameter_definition=BuiltInParameter.ALL_MODEL_MARK,
                                            parameter_value="",
                                        )
                                    except Exception as e:
                                        action_return_value.update_sep(
                                            False, "Failed with exception: {}".format(e)
                                        )
                                    return action_return_value

                                # set up the revit transaction
                                transaction = Transaction(
                                    doc, "Updating mark value to empty"
                                )
                                # run the action in the transaction
                                result = in_transaction_with_failure_handling(
                                    transaction=transaction,
                                    action=action,
                                    failure_config=failure_handling_config,
                                )
                                # setup return object message
                                message_prefix = "Updated mark on: {} [id:{}]".format(
                                    Element.Name.GetValue(element),
                                    element.Id.IntegerValue,
                                )
                                return_value.update_sep(
                                    result.status,
                                    "{} : {}".format(message_prefix, result.message),
                                )
                            else:
                                return_value.update_sep(
                                    True,
                                    "{}:  Element has no mark value: {}".format(
                                        self.filter_name,
                                        Element.Name.GetValue(element),
                                    ),
                                )
                        except Exception as e:
                            return_value.update_sep(
                                False,
                                "{}: Failed to solve warning duplicate mark with exception: {}".format(
                                    self.filter_name, e
                                ),
                            )
                    else:
                        return_value.update_sep(
                            True,
                            "{}: Element removed by filter: {}".format(
                                self.filter_name, Element.Name.GetValue(element)
                            ),
                        )

                # increase progress counter
                counter = counter + 1

                # check if cancelled
                if self.callback:
                    if self.callback.is_cancelled():
                        return_value.append_message("User cancelled!")
                        break
        else:
            return_value.update_sep(
                True,
                "{}: No warnings of type: duplicate mark in model.".format(
                    self.filter_name
                ),
            )
        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around Revit doors.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common.common import get_ids_from_element_collector

from Autodesk.Revit.DB import (
    BuiltInCategory,
    ElementCategoryFilter,
    FamilyInstance,
    FamilySymbol,
    FilteredElementCollector,
)

# ---------------- generic door collector functions ------------------------


def get_door_instances(doc):
    """
    Retrieves a list of door instances in a Revit model.

    :param doc: The Revit document object.
    :type doc: Document
    :return: A list of door instances in the Revit model.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """
    filter = ElementCategoryFilter(BuiltInCategory.OST_Doors)
    door_instances_in_model = (
        FilteredElementCollector(doc).OfClass(FamilyInstance).WherePasses(filter)
    )
    return door_instances_in_model


def get_door_symbols(doc):
    """
    Retrieves a list of door symbols in a Revit model.

    :param doc: The Revit document object.
    :type doc: Revit Document
    :return: A list of door symbols in the Revit model.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    filter = ElementCategoryFilter(BuiltInCategory.OST_Doors)
    door_symbols_in_model = (
        FilteredElementCollector(doc).OfClass(FamilySymbol).WherePasses(filter)
    )
    return door_symbols_in_model


def get_door_families(doc):
    """
    Retrieves a list of door families in a Revit model.

    :param doc: The Revit document object.
    :type doc: Revit Document
    :return: A list of door families in the Revit model.
    :rtype: list
    """

    door_families = []
    door_family_ids = []
    door_symbols = get_door_symbols(doc=doc)
    for door_symbol in door_symbols:
        if door_symbol.Family.Id not in door_family_ids:
            door_family_ids.append(door_symbol.Family.Id)
            door_families.append(door_symbol.Family)
    return door_families


# -------------------------- curtain wall door collector functions -----------------------------


def get_curtain_wall_door_instances(doc):
    """
    Retrieves a list of door instances in a Revit model that are considered curtain wall doors.

    :param doc: The Revit document object. It represents the Revit model.
    :type doc: Revit Document
    :return: A list of door instances in the Revit model that are considered curtain wall doors.
    :rtype: list of Autodesk.Revit.DB.FamilyInstance
    """

    curtain_wall_door_instances = []
    curtain_wall_door_symbols = get_curtain_wall_door_symbols(doc=doc)
    curtain_wall_door_symbol_ids = get_ids_from_element_collector(
        curtain_wall_door_symbols
    )
    door_instances = get_door_instances(doc=doc)

    # loop over instances and check which symbol is a curtain wall door symbol
    for door_instance in door_instances:
        if door_instance.Symbol.Id in curtain_wall_door_symbol_ids:
            curtain_wall_door_instances.append(door_instance)

    return curtain_wall_door_instances


def get_curtain_wall_door_symbols(doc):
    """
    Retrieves a list of door symbols that are considered curtain wall doors in a Revit model.

    :param doc: The Revit document object.
    :type doc: Revit Document
    :return: A list of door symbols that are considered curtain wall doors in the Revit model.
    :rtype: list
    """

    curtain_wall_door_symbols = []

    # get all door symbols in model
    door_symbols_in_model = get_door_symbols(doc)

    # get all panel symbols in model
    filter_panels = ElementCategoryFilter(BuiltInCategory.OST_CurtainWallPanels)
    panels_in_model = (
        FilteredElementCollector(doc).OfClass(FamilySymbol).WherePasses(filter_panels)
    )

    # convert panel symbols to ids
    panel_ids = get_ids_from_element_collector(col=panels_in_model)

    # check if similar type query returns any panel id...if so then this is a curtain wall door
    for door_symbol in door_symbols_in_model:
        similar_type_ids = door_symbol.GetSimilarTypes()
        for panel_id in panel_ids:
            if panel_id in similar_type_ids:
                # found a curtain wall door
                curtain_wall_door_symbols.append(door_symbol)

    return curtain_wall_door_symbols


def get_curtain_wall_door_families(doc):
    """
    Retrieves a list of curtain wall door families in a Revit model.

    :param doc: The Revit document object.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of curtain wall door families in the Revit model.
    :rtype: list[Autodesk.Revit.DB.Family]
    """

    curtain_wall_door_families = []
    curtain_wall_door_family_ids = []

    # get all curtain wall door symbols in the model
    curtain_wall_door_symbols = get_curtain_wall_door_symbols(doc=doc)

    # loop over symbols and get distinct list of families
    for curtain_wall_door_symbol in curtain_wall_door_symbols:
        if curtain_wall_door_symbol.Family.Id not in curtain_wall_door_family_ids:
            curtain_wall_door_family_ids.append(curtain_wall_door_symbol.Family.Id)
            curtain_wall_door_families.append(curtain_wall_door_symbol.Family)
    return curtain_wall_door_families


# -------------------------- basic wall door collector functions -----------------------------


def get_basic_wall_door_instances(doc):
    """
    Retrieves a list of door instances in a Revit model that are not considered curtain wall doors.

    :param doc: The Revit document object.
    :type doc: Revit Document
    :return: A list of door instances in the Revit model that are not considered curtain wall doors.
    :rtype: list
    """

    basic_wall_door_instances = []
    curtain_wall_door_symbols = get_curtain_wall_door_symbols(doc=doc)
    curtain_wall_door_symbol_ids = get_ids_from_element_collector(
        curtain_wall_door_symbols
    )
    door_instances = get_door_instances(doc=doc)

    # loop over instances and check which symbol is not a curtain wall door symbol
    for door_instance in door_instances:
        if door_instance.Symbol.Id not in curtain_wall_door_symbol_ids:
            basic_wall_door_instances.append(door_instance)

    return basic_wall_door_instances


def get_basic_wall_door_symbols(doc):
    """
    Returns a list of door symbols that are not curtain wall doors in a Revit model.

    :param doc: The Revit document object.
    :type doc: Document
    :return: A list of door symbols that are not curtain wall doors in the Revit model.
    :rtype: list[FamilySymbol]
    """

    basic_wall_door_symbols = []
    # get all door symbols in model
    door_symbols_in_model = get_curtain_wall_door_symbols(doc=doc)
    # get all curtain wall door symbols
    curtain_wall_door_symbols = get_curtain_wall_door_symbols(doc=doc)
    curtain_wall_door_symbol_ids = get_ids_from_element_collector(
        curtain_wall_door_symbols
    )
    # difference between curtain wall doors and all doors are basic wall door symbols
    for door_symbol in door_symbols_in_model:
        if door_symbol.Id not in curtain_wall_door_symbol_ids:
            basic_wall_door_symbols.append(door_symbol)

    return basic_wall_door_symbols


def get_basic_wall_door_families(doc):
    """
    Retrieves a list of basic wall door families in a Revit model.

    :param doc: The Revit document object.
    :type doc: Document
    :return: A list of basic wall door families in the Revit model.
    :rtype: list
    """

    basic_wall_door_families = []
    basic_wall_door_family_ids = []

    # get all basic wall door symbols in the model
    basic_wall_door_symbols = get_basic_wall_door_symbols(doc=doc)

    # loop over symbols and get distinct list of families
    for basic_wall_door_symbol in basic_wall_door_symbols:
        if basic_wall_door_symbol.Family.Id not in basic_wall_door_family_ids:
            basic_wall_door_family_ids.append(basic_wall_door_symbol.Family.Id)
            basic_wall_door_families.append(basic_wall_door_symbol.Family)
    return basic_wall_door_families



# -*- coding: utf-8 -*-
"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Class for family parameter data storage class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Used to store values of family parameters by family type.

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# note this module is not utf-8 encoded as it contains special characters representing units of measurement that are not utf-8 encoded 


from duHast.Utilities.Objects.base import Base
from duHast.Revit.Family.Data.Objects import ifamily_data_storage as IFamDataStorage


# map the storage type as reported in the part atom export to the storage type used in the catalogue file header row
PARAMETER_STORAGE_TYPE_MAPPER = {
    "Angle":"ANGLE",
    "Area":"AREA",
    "Cost per Area":"COST_PER_AREA",
    "Currency":"CURRENCY",
    "Distance":"DISTANCE",
    "Fill Pattern":"OTHER",
    "Image":"OTHER",
    "Length":"LENGTH",
    "Material":"OTHER",
    "Multiline Text":"OTHER",
    "Number":"OTHER",
    "Rotation Angle":"ROTATION_ANGLE",
    "Slope":"SLOPE",
    "Speed":"SPEED",
    "Time":"TIMEINTERVAL",
    "Text":"OTHER",
    "URL":"OTHER",
    "Volume":"VOLUME",
    "Yes/No":"OTHER",
}

# map the units as reported in the part atom export to the units used in the catalogue file header row (depending on the storage type)
PARAMETER_UNITS_MAPPER = {
    b"mm":[["Length","MILLIMETERS"], ["Distance","MILLIMETERS"]],
    b"cm":[["Length","CENTIMETERS"], ["Distance","CENTIMETERS"]],
    b"m":[["Length","METERS"], ["Distance","METERS"]],
    b"mm\xc2\xb2":[["Area","SQUARE_MILLIMETERS"]],
    b"cm\xc2\xb2":[["Area","SQUARE_CENTIMETERS"]],
    b"m\xc2\xb2":[["Area","SQUARE_METERS"]],
    b"mm\xc2\xb3":[["Volume","CUBIC_MILLIMETERS"]],
    b"cm\xc2\xb3":[["Volume","CUBIC_CENTIMETERS"]],
    b"m\xc2\xb3":[["Volume","CUBIC_METERS"]],
    b"m/s":[["Speed","METERS_PER_SECOND"]],
    b"cm/m":[["Speed","CENTIMETERS_PER_MINUTE"]],
    b"km/h":[["Speed","KILOMETERS_PER_HOUR"]],
    b"ms":[["Time","MILLISECONDS"]],
    b"s":[["Time","SECONDS"]],
    b"min":[["Time","MINUTES"]],
    b"h":[["Time","HOURS"]],
    b"grad":[["Rotation Angle","GRADIANS"],["Angle","GRADIANS"]],
    b"rad":[["Angle","RADIANS"],["Rotation Angle","RADIANS"]],
    b"\xc2\xb0":[["Slope","SLOPE_DEGREES"],["Rotation Angle","DEGREES"],["Angle","DEGREES"]],
    b"$":[["Currency","CURRENCY"]],
    b"$/m\xc2\xb2":[["Cost per Area", "COST_PER_SQUARE_METER"]],
    b"unitless":[["",""]],
}

class FamilyTypeParameterDataStorage(IFamDataStorage.IFamilyDataStorage):

    # data type for this class ( used in reports as first entry per row )
    data_type = "FamilyTypeParameter"

    unit_type_compare_values_as_floats = [
        "Length",
        "Area",
        "Angle",
        "Currency",
        "Flow",
    ]

    # number of properties in this class ( used in report reader function )
    number_of_properties = 10

    def __init__(self, 
            root_name_path,
            root_category_path,
            family_name,
            family_file_path,
            family_type_name,
            name, 
            type, 
            type_of_parameter, 
            units, 
            value
        ):
        """
        constructor

        :param root_name_path: root name path (i.e. rootFamilyName :: nestedFamilyNameOne :: nestedFamilyTwo)
        :type root_name_path: str
        :param root_category_path: root category path (i.e. rootCategoryName :: nestedCategoryNameOne :: nestedCategoryTwo)
        :type root_category_path: str
        :param family_name: name of the family
        :type family_name: str
        :param family_file_path: file path of the family
        :type family_file_path: str
        :param name: name of the family parameter
        :param type: type of the parameter ( i.e. shared, system, custom)
        :param type_of_parameter: unit type of the parameter ( i.e. length, area, volume, string, etc.)
        :param units: units of the parameter, i.e. mm ( there are parameter type which do not have units i.e. string)
        :param value: value of the parameter
        """

        super(FamilyTypeParameterDataStorage, self).__init__(
            data_type=FamilyTypeParameterDataStorage.data_type,
            root_name_path=root_name_path,
            root_category_path=root_category_path,
            family_name=family_name,
            family_file_path=family_file_path,
        )

        self.family_type_name = family_type_name  # name of the family type
        self.name = name  # name of the family type
        self.type = type  # type of the parameter ( i.e. shared, system, custom)
        self.type_of_parameter = type_of_parameter  # unit type of the parameter ( i.e. length, area, volume, string, etc.)
        self.units = units  # units of the parameter ( there are parameter type which do not have units i.e. string)
        self.value = value  # value of the parameter
        self.value_as_float = None

        # attempt to convert value to float if possible
        if self.type_of_parameter in self.unit_type_compare_values_as_floats:
            try:
                self.value_as_float = float(self.value)
            except Exception:
                self.value_as_float = None
          

    def __eq__(self, other):
        """
        equal compare

        :param other: object to compare with
        :return: True if equal, False if not
        """

        if not isinstance(other, FamilyTypeParameterDataStorage):
            raise ValueError("other must be an instance of FamilyTypeParameterDataStorage")
        
        # compare values as floats if the type of parameter is in the list of unit types
        # that should be compared as floats and the value as float is not None
        if self.type_of_parameter in self.unit_type_compare_values_as_floats and self.value_as_float is not None:
            return (
                self.data_type == other.data_type and
                self.name == other.name and
                self.type == other.type and
                self.type_of_parameter == other.type_of_parameter and
                self.units == other.units and
                self.value_as_float == other.value_as_float
            )
        else:
            # compare values as strings
            return (
                self.data_type == other.data_type and
                self.name == other.name and
                self.type == other.type and
                self.type_of_parameter == other.type_of_parameter and
                self.units == other.units and
                self.value == other.value
            )

    def __ne__(self, other):
        """
        not equal compare

        :param other: object to compare with
        :return: True if not equal, False if equal
        """
        return not self.__eq__(other)

    def __hash__(self):
        """
        hash function for this class

        :return: hash value
        """

        return hash(
            (
                self.data_type,
                self.name,
                self.type,
                self.type_of_parameter,
                self.units,
                self.value,
                self.value_as_float,
            )
        )

    def get_difference(self, other):
        """
        get the difference between this object and another object

        :param other: object to compare with
        :return: a list of differences
        :rtype: [str]
        """

        if not isinstance(other, FamilyTypeParameterDataStorage):
            return NotImplemented

        differences = []

        if self.name != other.name:
            differences.append("name: {} != {}".format(self.name, other.name))
        if self.type != other.type:
            differences.append("type: {} != {}".format(self.type, other.type))
        if self.type_of_parameter != other.type_of_parameter:
            differences.append(
                "type_of_parameter: {} != {}".format(
                    self.type_of_parameter, other.type_of_parameter
                )
            )
        if self.units != other.units:
            differences.append("units: {} != {}".format(self.units, other.units))
        
        # check if values need to be compared as floats
        if self.type_of_parameter in self.unit_type_compare_values_as_floats and self.value_as_float is not None and other.value_as_float is not None:
            # compare values as floats
            if self.value_as_float != other.value_as_float:
                differences.append("value: {} != {}".format(self.value_as_float, other.value_as_float))
        else:
            # compare values as strings
            if self.value != other.value:
                differences.append("value: {} != {}".format(self.value, other.value))

        return differences

    def get_report_data(self):
        """
        get the report data for this object

        :return: the report data
        :rtype: [str]
        """

        return [
            self.name,
            self.type,
            self.type_of_parameter,
            self.units,
            self.value,
        ]
    

    def get_catalogue_file_data(self):
        """
        Get the catalogue file data for this parameter ( the value of the parameter)

        :return: the catalogue file data
        :rtype: [str]
        """

        return self.value
    

    def get_catalogue_file_header_row(self):
        """
        Get the catalogue file header row for this object sample ParameterName##LENGTH##MILLIMETERS

        :return: the catalogue file header row entry for this parameter
        :rtype: str
        """

        
        # check if the type of parameter is in the mapper
        if self.type_of_parameter not in PARAMETER_STORAGE_TYPE_MAPPER:
            raise ValueError("Unknown parameter type: {} for parameter: {}".format(self.type_of_parameter, self.name))
        
        # get the mapper name for the type of parameter
        type_name_revised_for_file = PARAMETER_STORAGE_TYPE_MAPPER[self.type_of_parameter]

        # check if the units are in the mapper
        # dictionary uses as keys byte strings due to non unicode characters!
        if self.units not in PARAMETER_UNITS_MAPPER:
            raise ValueError("Unknown parameter unit: {} for parameter: {}".format( self.units, self.name))
        
        # get the mapper name for the units
        unit_names_lists_for_catalogue_file = PARAMETER_UNITS_MAPPER[self.units]

        found_unit_match = False
        unit_names_for_catalogue_file=""
        # get the unit name depending on the storage type of parameter
        if len(unit_names_lists_for_catalogue_file) == 1:
            # if there is only one unit name for the storage type
            unit_names_for_catalogue_file = unit_names_lists_for_catalogue_file[0][1]
            found_unit_match = True
        else:
            # if there are multiple unit names depending on the storage type
            for unit_name_list in unit_names_lists_for_catalogue_file:
                if unit_name_list[0] == self.type_of_parameter:
                    unit_names_for_catalogue_file = unit_name_list[1]
                    found_unit_match = True
                    break

        # check if the unit name was found
        if not found_unit_match:
            raise ValueError("Unknown parameter unit: {} for parameter: {}".format(self.units, self.name))


        # return the catalogue file header row entry for this parameter
        return "{}##{}##{}".format(self.name, type_name_revised_for_file, unit_names_for_catalogue_file)


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A module to provide pyRevit multiselectors for some standard Revit elements.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from collections import defaultdict
from duHast.Revit.Levels.Objects.LevelBaseObject import LevelBaseObj
from pyrevit import forms
from duHast.Revit.Rooms.rooms import get_all_placed_rooms
from duHast.Revit.Rooms.room_common_parameters import get_room_number
from duHast.Revit.Rooms.Objects.RoomBaseObject import RoomBaseObj
from duHast.Revit.Levels.levels import get_levels_list_ascending

# ---------------------------------------------------------------------------- Rooms


def room_selector(rvt_doc, multiple=True):
    """
    pyRevit Select from list UI object to pick room or rooms.
    :param rvt_doc: Revit Document
    :type rvt_doc: Document
    :param multiple: Allow multiple selection
    :type multiple: bool
    :return: Room or list of rooms
    :rtype: Room or list[Room]
    """
    all_rooms = get_all_placed_rooms(rvt_doc)
    sorted_rooms = sorted(all_rooms, key=lambda rm: str(get_room_number(rm)))
    rooms_for_ui = defaultdict(list)

    for rm in sorted_rooms:
        obj = RoomBaseObj(rvt_doc, rm)
        rooms_for_ui[obj.level.Name].append(obj)
        rooms_for_ui["All levels"].append(obj)

    btn_name = "Select Rooms" if multiple else "Select Room"

    select_room_ui = forms.SelectFromList.show(
        rooms_for_ui,
        multiselect=multiple,
        group_selector_title="Filter by level:",
        name_attr="number_name_comb",
        button_name=btn_name,
    )

    if select_room_ui:
        if multiple:
            return [x.room for x in select_room_ui]
        else:
            return select_room_ui.room

    return None


def pick_room(rvt_doc):
    """
    PyRevit multiselector to pick a single room.
    """
    return room_selector(rvt_doc, multiple=False)


def pick_rooms(rvt_doc):
    """
    PyRevit multiselector to pick multiple rooms.
    """
    return room_selector(rvt_doc, multiple=True)


# ---------------------------------------------------------------------------- Levels


def level_selector(rvt_doc, multiple=True):
    lvls = get_levels_list_ascending(rvt_doc)
    lvl_objs = [LevelBaseObj(lvl) for lvl in lvls]
    btn_name = "Select Levels" if multiple else "Select Level"

    select_level_ui = forms.SelectFromList.show(
        lvl_objs,
        multiselect=multiple,
        name_attr="name_elev_comb",
        button_name=btn_name,
    )

    if select_level_ui:
        if multiple:
            lvl_list = [x.level for x in select_level_ui]
        else:
            lvl_list = select_level_ui.level

        return lvl_list

    return None


def pick_level(rvt_doc):
    """
    PyRevit multiselector to pick a single level.
    """
    return level_selector(rvt_doc, multiple=False)


def pick_levels(rvt_doc):
    """
    PyRevit multiselector to pick multiple levels.
    """
    return level_selector(rvt_doc, multiple=True)



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family report data utility module containing functions to read the data from file.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Reads family base data into two list of named tuples.

rootFamily:

- name 
- category 
- filePath 


nestedFamily:

- name 
- category 
- filePath 
- rootPath  [str]
- categoryPath [str]

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

from collections import namedtuple

from duHast.Utilities.Objects import result as res
from duHast.Utilities import files_csv as fileCSV, files_io as fileIO
from duHast.Revit.Family.Data.family_base_data_utils_deprecated import (
    read_overall_family_data_list,
)

# tuples containing base family data read from file
rootFamily = namedtuple("rootFamily", "name category filePath")
nestedFamily = namedtuple(
    "nestedFamily", "name category filePath rootPath categoryPath"
)

# row structure of report data file
BASE_DATA_LIST_INDEX_ROOT_PATH = 0
BASE_DATA_LIST_INDEX_ROOT_CATEGORY_PATH = 1
BASE_DATA_LIST_INDEX_FAMILY_NAME = 2
BASE_DATA_LIST_INDEX_FAMILY_FILE_PATH = 3

# exceptions
EXCEPTION_NO_FAMILY_BASE_DATA_FILES = "Report data list files do not exist."
EXCEPTION_EMPTY_FAMILY_BASE_DATA_FILES = "Empty report data list file!"

# ------------------------------------- combining reports --------------------------------------------


def _get_data_rows_from_dictionary(dic):
    """
    Builds list of data rows from dictionary past in

    :param dic: Dictionary where key is a tuple and values is a list of list of strings
    :type dic: {named tuple: [[str]]}
    :param dataList: List of list of strings
    :type dataList: [[str]]

    :return: List of list of strings
    :rtype: [[str]]
    """
    print(dic)
    dataList = []
    # get rows from dictionary
    for k, v in dic.items():
        # get data rows for root family
        for rootData in v[0]:
            dataList.append(rootData.report_data)
        # get data rows for any nested families
        for nestedFamRowValue in v[1]:
            dataList.append(nestedFamRowValue.report_data)
    return dataList


def _compare_family_dictionaries(previous_aggregated_data, new_aggregated_data):
    """
    Compares two aggregate data dictionaries. Any new root family from newAgData ( root family occurring in newAgData only) will be add to the previousAgData dictionary.
    Any existing root family (root family occurring in previous and new aggregate data dictionaries) will be updated in the previousAgData dictionary with row data from the newAgData data dictionary.


    :param previous_aggregated_data: A dictionary containing aggregated family data from the previous report.
    :type previous_aggregated_data: {key:str, value ([str],[str])}
    :param new_aggregated_data: A dictionary containing aggregated family data from the new report.
    :type new_aggregated_data: {key:str, value ([str],[str])}
    :return:

        If previous_aggregated_data is empty and new_aggregated_data contains data, new_aggregated_data will be returned unchanged.
        If new_aggregated_data is empty and previous_aggregated_data contains data, previous_aggregated_data will be returned unchanged.
        if both dictionary are empty an empty dictionary will be returned.

    :rtype: {key:str, value ([str],[str])}
    """

    return_value = res.Result()
    # check corner cases:
    if len(new_aggregated_data) == 0 and len(previous_aggregated_data) > 0:
        # new is empty, but previous has data
        return_value.update_sep(
            True, "New report data is empty, using previous report data only"
        )
        return_value.result.append(previous_aggregated_data)
    elif len(new_aggregated_data) > 0 and len(previous_aggregated_data) == 0:
        # new has data, but previous is empty
        return_value.update_sep(
            True, "Previous report data is empty, using new report data only"
        )
        return_value.result.append(new_aggregated_data)
    elif len(new_aggregated_data) == 0 and len(previous_aggregated_data) == 0:
        # new is empty, previous is empty
        return_value.update_sep(
            True, "Previous report data and new report data are empty!"
        )
        return_value.result.append({})
    else:
        # other and current have data
        for family_path, family_data in new_aggregated_data.items():
            if family_path in previous_aggregated_data:
                return_value.append_message(
                    "Substituting family data: {}".format(family_path)
                )
            else:
                return_value.append_message(
                    "Adding new family data: {}".format(family_path)
                )
            previous_aggregated_data[family_path] = new_aggregated_data[family_path]
        return_value.result.append(previous_aggregated_data)
    return return_value


def _get_nested_families_belonging_to_root_families(root_family, nested_families):
    """
    Returns a list of all row data of nested families belonging to a given root family.

    :param rootFam: A tuple of a root family from a report.
    :type rootFam: tuple of type 'rootFamily'
    :param nestedFamilies: A list of tuples of all nested families in a report
    :type nestedFamilies: [tuple of type 'nestedFamily']

    :return:
    :rtype: _type_
    """

    nested_families_belonging_to_root_families = []
    for nested_family in nested_families:
        if (
            root_family.name == nested_family.rootPath[0]
            and root_family.category == nested_family.categoryPath[0]
        ):
            nested_families_belonging_to_root_families.append(
                nested_families[nested_family][0]
            )
            print("nested family in list", nested_families[nested_family][0])
        else:
            print(
                "root family name: [{}] root family category: [{}] nested family root path: [{}] nested family category path: [{}]".format(
                    root_family.name,
                    root_family.category,
                    nested_family.rootPath[0],
                    nested_family.categoryPath[0],
                )
            )
    return nested_families_belonging_to_root_families


def _aggregate_family_data(root_families, nested_families):
    """
    Returns a dictionary where key are all the root family file path from a report and value is a tuple of two list of strings containing
    the row data read from report file for the root family itself (first list) and the row data read from report file for any nested families (second list).

    :param rootFamilies: A list containing tuples of all root families in a report.
    :type rootFamilies: [tuple of type 'rootFamily']
    :param nestedFamilies: A list of tuples of all nested families in a report
    :type nestedFamilies: [tuple of type 'nestedFamily']

    :return: Returns a dictionary where key is the root family file path and value is a tuple of root family at index zero and nested families at index 1
    :rtype: {key:str, value (root family,[nested families])}
    """

    # key is root family, value is tuple of csv row representing the root family data and list of rows each representing a nested family data
    aggregated_family_data = {}
    for root_family in root_families:
        nested_families_of_root_family_row_data = (
            _get_nested_families_belonging_to_root_families(
                root_family, nested_families
            )
        )
        # key is the unique family file path of the root family
        # value is a tuple of two lists : root  at index 0, nested fam at index 1
        aggregated_family_data[root_family.filePath] = (
            root_family,
            nested_families_of_root_family_row_data,
        )
    return aggregated_family_data


def _check_families_still_exist(family_data):
    """
    Checks whether families still exist on file server.

    Reason why families no longer exist:

    - family got deleted or moved
    - family got renamed

    :param famData: A dictionary containing aggregated family data from the a report.
    :type famData: {key:str, value ([str],[str])}

    :return:
        Result class instance.

        - .status True if successfully removed any outdated family data or None needed removing. Otherwise False.
        - .message will contain list of families removed or message nothing needed to be removed.
        - . result will contain past in dictionary at index 0

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        remove_keys = []
        # get keys from dic as a list
        # check which ones do not exist anymore
        for file_path in family_data.keys():
            if fileIO.file_exist(file_path) == False:
                remove_keys.append(file_path)

        # check if any family requires to be removed from the data set
        if len(remove_keys) > 0:
            # remove those keys from dictionary
            for key in remove_keys:
                remove_single_key = family_data.pop(key, None)
                if remove_single_key != None:
                    return_value.append_message(
                        "Removed family from data: {}".format(key)
                    )
                else:
                    return_value.append_message(
                        "Failed to removed family from data: {}".format(key)
                    )
        else:
            return_value.append_message("No family required removing from data.")

        # update return data
        return_value.update_sep(True, "Successfully updated family data.")
        return_value.result.append(family_data)

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to check whether families still exist with exception: {}".format(e),
        )
    return return_value


def combine_reports(previous_report_path, new_report_path):
    """
    This combines two reports by:

    - building an aggregate data dictionary of each report (key root family file path, values lists containing the row data read from file for the root family as well as any nested families)
    - comparing the previous report dictionary with the new report dictionary and
        - adding any new families found in the new report dictionary
        - updating any previous report families found with data matching the root family in the new report dictionary

    All reports start with the following 2 columns:
    root	rootCategory

    First entry (after split at separator) in each of these columns identifies root family uniquely.
    Assume that new report only ever adds or substitutes entries in previous report but does not delete from it!

    This function checks at the end whether families still exist on file server. If not, they will be removed from the data set.

    :param previous_report_path: A fully qualified file path to the previous report file.
    :type previous_report_path: str
    :param new_report_path: A fully qualified file path to the new report file.
    :type new_report_path: str

    :return: list of lists of report rows
    :rtype: [[str]]
    """

    return_value = res.Result()
    # read families from both reports
    # ...compare them:
    # take all families from current report and all none matching families from the other report

    previous_aggregated_families = {}
    new_aggregated_families = {}

    print("reading previous")
    # previous report
    try:
        previous_root, previous_nested = read_overall_family_data_list(
            previous_report_path
        )
        # previous_root, previous_nested = read_unique_families_with_row_data_from_report(
        #    previous_report_path
        # )
        return_value.append_message(
            "Previous report: found {} root families.".format(len(previous_root))
        )
        return_value.append_message(
            "Previous report: found {} nested families.".format(len(previous_nested))
        )
        # build dictionary containing all family data per root family
        previous_aggregated_families = _aggregate_family_data(
            previous_root, previous_nested
        )
    except Exception as e:
        # check whether empty file exception
        if str(e) != EXCEPTION_EMPTY_FAMILY_BASE_DATA_FILES:
            raise e

    print("reading new")
    # new report
    try:
        new_root, new_nested = read_overall_family_data_list(new_report_path)
        # new_root, new_nested = read_unique_families_with_row_data_from_report(
        #    new_report_path
        # )
        # build dictionary containing all family data per root family
        new_aggregated_families = _aggregate_family_data(new_root, new_nested)
        return_value.append_message(
            "New report: found {} root families.".format(len(new_root)) + ""
        )
        return_value.append_message(
            "New report: found {} nested families.".format(len(new_nested)) + ""
        )
    except Exception as e:
        # check whether empty file exception
        if str(e) != EXCEPTION_EMPTY_FAMILY_BASE_DATA_FILES:
            raise e

    print("comparing")
    # compare dictionaries: build unique list of families
    unique_family_data_status = _compare_family_dictionaries(
        previous_aggregated_families, new_aggregated_families
    )
    return_value.update(unique_family_data_status)
    unique_family_data = unique_family_data_status.result[0]

    print("checking")
    # check whether families still exist on file server
    remove_none_existing_families = _check_families_still_exist(unique_family_data)
    return_value.update(remove_none_existing_families)
    # only update family data if culling occurred without any exceptions
    if remove_none_existing_families.status:
        unique_family_data = remove_none_existing_families.result[0]

    # get report header row (there should be a previous report file...otherwise this will write an empty header row)
    header_row_result = fileCSV.get_first_row_in_csv_file(previous_report_path)
    if header_row_result.status == False:
        return_value.update_sep(
            False, "Failed to read header row from previous report file."
        )
        print("failed to read header row from previous report file.")
        return return_value
    
    header_row = header_row_result.result
    
    # header_row = header.split(",")
    print("building")
    # build list of data rows
    rows_current = _get_data_rows_from_dictionary(unique_family_data)
    # sort rows by root ( first entry ) since other code (circ reference checker for instance) expects data sorted
    rows_current.sort()
    # start with header row
    rows_current.insert(0, header_row)
    # overwrite return result value since it is already containing data from previous operations
    return_value.result = rows_current
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit category report functions .
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.LinePattern.line_patterns import (
    get_line_pattern_from_category,
    PROPERTY_PATTERN_NAME,
    PROPERTY_PATTERN_ID,
)

from duHast.Revit.Categories.categories import (
    get_family_category,
    get_main_sub_categories,
    get_other_sub_categories,
)
from duHast.Revit.Categories.Utility.category_properties_get_utils import (
    get_category_colour,
    get_category_line_weights,
    get_category_material,
)
from duHast.Revit.Categories.Utility.category_property_names import (
    CATEGORY_GRAPHIC_STYLE_3D,
    CATEGORY_GRAPHIC_STYLE_CUT,
    CATEGORY_GRAPHIC_STYLE_PROJECTION,
    PROPERTY_LINE_COLOUR_BLUE_NAME,
    PROPERTY_LINE_COLOUR_GREEN_NAME,
    PROPERTY_LINE_COLOUR_RED_NAME,
    PROPERTY_LINE_WEIGHT_CUT_NAME,
    PROPERTY_LINE_WEIGHT_PROJECTION_NAME,
    PROPERTY_MATERIAL_ID,
    PROPERTY_MATERIAL_NAME,
)
from duHast.Revit.Categories.Utility.elements_by_category_utils import (
    get_elements_by_category,
)

from duHast.Utilities.utility import encode_utf8


def build_report_data_by_category(doc, dic, family_cat, main_cat_name, doc_file_path):
    """
    Formats category properties into lists for reports
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param dic: dictionary containing category properties
    :type dic: _type_
    :param family_cat: The family category name.
    :type family_cat: str
    :param main_cat_name: A hard coded revit category name. Can be the same as familyCat.
    :type main_cat_name: str
    :param doc_file_path: The fully qualified family file path.
    :type doc_file_path: str
    :return: A list of list of strings. Each row represents one category.
    :rtype: list[list[str]]
    """

    data = []
    for key in dic:
        row = [
            encode_utf8(str(doc_file_path)),
            encode_utf8(family_cat),
            encode_utf8(main_cat_name),
            encode_utf8(key),
            str(dic[key].Id),
        ]
        # get elements
        elements = get_elements_by_category(doc, dic[key])
        # get properties
        dic_material = get_category_material(dic[key])
        row.append(encode_utf8(str(dic_material[PROPERTY_MATERIAL_NAME])))
        row.append(encode_utf8(str(dic_material[PROPERTY_MATERIAL_ID])))
        # line pattern
        dic_pattern = get_line_pattern_from_category(dic[key], doc)
        row.append(encode_utf8(str(dic_pattern[PROPERTY_PATTERN_NAME])))
        row.append(encode_utf8(str(dic_pattern[PROPERTY_PATTERN_ID])))
        # line weights
        dic_line_weights = get_category_line_weights(dic[key])
        row.append(
            encode_utf8(str(dic_line_weights[PROPERTY_LINE_WEIGHT_PROJECTION_NAME]))
        )
        row.append(encode_utf8(str(dic_line_weights[PROPERTY_LINE_WEIGHT_CUT_NAME])))
        # category colour
        dic_colour = get_category_colour(dic[key])
        row.append(encode_utf8(str(dic_colour[PROPERTY_LINE_COLOUR_RED_NAME])))
        row.append(encode_utf8(str(dic_colour[PROPERTY_LINE_COLOUR_GREEN_NAME])))
        row.append(encode_utf8(str(dic_colour[PROPERTY_LINE_COLOUR_BLUE_NAME])))
        # elements
        row.append(encode_utf8(str(len(elements[CATEGORY_GRAPHIC_STYLE_3D]))))
        row.append(encode_utf8(str(len(elements[CATEGORY_GRAPHIC_STYLE_PROJECTION]))))
        row.append(encode_utf8(str(len(elements[CATEGORY_GRAPHIC_STYLE_CUT]))))

        data.append(row)
    return data


def get_report_data(doc, revit_file_path):
    """
    Reports all categories, their properties and all elements belonging to them.
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: The fully qualified family file path.
    :type revit_file_path: str
    :return: A list of list of strings. Each row represents one category.
    :rtype: list[list[str]]
    """

    data = []
    # get all sub categories in family and associates elements;
    sub_cats = get_main_sub_categories(
        doc
    )  # i/e family is specialty equipment and all its associated sub categories
    family_cat = get_family_category(
        doc
    )  # any 3D element which is set to 'None' in subcategory (if family is specialty equipment so is this element)
    other_cats = get_other_sub_categories(doc)  # Imports in Families cats are here
    family_cat_name = list(family_cat.keys())[0]
    # build output
    data = build_report_data_by_category(
        doc, family_cat, family_cat_name, family_cat_name, revit_file_path
    )
    data = data + build_report_data_by_category(
        doc, sub_cats, family_cat_name, family_cat_name, revit_file_path
    )
    # check for imports
    if "Imports in Families" in other_cats:
        data = data + build_report_data_by_category(
            doc,
            other_cats["Imports in Families"],
            family_cat_name,
            "Imports in Families",
            revit_file_path,
        )
    return data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A module with helper function around family types data extraction using Revit xml export functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Supports 2 methods of data extraction:

- from family file on disk
- from family element instance in document

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import tempfile
import os

from Autodesk.Revit.DB import Element

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Family.Utility.xml_family_type_reader import read_xml_into_storage
from duHast.Revit.Family.family_functions import get_symbol_names_of_family
from duHast.Revit.Family.Utility.xml_create_atom_exports import write_data_to_temp_xml_file_and_read_it_back


def get_type_data_via_XML_from_family_object(revit_family, family_name_nesting_path = None, root_category_path = None):
    """
    Get the family type data from the family element in a REvit document using the XML extraction method.

    :param revit_family: The Revit family object.
    :type revit_family: Autodesk.Revit.DB.Family
    :param family_name_nesting_path: The nesting path of the family in a tree: rootFamilyName :: nestedFamilyNameOne :: nestedFamilyTwo\
        This includes the actual family name as the last node.
    :type family_name_nesting_path: str
    :param root_category_path: The path of the family category in a tree: rootCategoryName :: nestedCategoryNameOne :: nestedCategoryTwo\
        This includes the actual category name as the last node.

    :return: A result object with .result containing a single FamilyTypeDataStorage object. (or empty if failed)
    :rtype: Result
    """

    return_value = Result()
    try:
        # set up action to write xml data
        def action(temp_path_xml):
            action_return_value = Result()
            try:
                # Save XML file to temporary location
                revit_family.ExtractPartAtom(temp_path_xml)
                action_return_value.update_sep(True, "Wrote data to XML file")
            except Exception as e:
                action_return_value.update_sep(
                    False, "Failed to write XML data: {}".format(e)
                )
            return action_return_value

        # Write the data to an XML file and read it back
        doc_xml_result = write_data_to_temp_xml_file_and_read_it_back(action)
        return_value.update(doc_xml_result)

        # check if an xml document was created
        if doc_xml_result.status is False:
            return return_value

        # set the family name and root category path
        fam_name = Element.Name.GetValue(revit_family) if family_name_nesting_path is None else family_name_nesting_path
        root_category_path = "None" if root_category_path is None else root_category_path

        # read the xml data into the storage object
        type_data = read_xml_into_storage(
            doc_xml_result.result,
            family_name=fam_name,
            family_path="",
            root_category_path=root_category_path,
        )

        # it looks like the part atom extraction does sometime include types which are no longer present
        # in the family document (ghost types) so we need to check if we have any types
        # since family inherits from Element, I should be able to get the document from the family object
        # and check if the types are still present in the document
        symbol_names = get_symbol_names_of_family(revit_family)
        type_data.remove_ghost_types(symbol_names)
    
        # store list in return object ( clear any previous results )
        return_value.result = [type_data]
    except Exception as e:
        return_value.update_sep(False, "{}".format(e))

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit curve loops helper functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from System.Collections.Generic import List

from duHast.Revit.Common.transaction import in_transaction
from duHast.Utilities.Objects.result import Result
from duHast.Utilities.unit_conversion import  convert_mm_to_imperial_feet
from duHast.Revit.DetailItems.filled_regions_create import create_filled_region_by_view
from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value, getter_double_as_double_converted_to_metric
from duHast.Revit.Common.delete import delete_by_element_ids

# import Autodesk
from Autodesk.Revit.DB import (
    Arc,
    BuiltInParameter,
    Curve,
    CurveArray,
    CurveArrArray,
    CurveLoop,
    Transaction,
    Transform,
    XYZ,
)

def create_curve_loops_through_transform(curve_loops, transform, convert_net_list=False):
    """
    Creates a new curve loop by transforming the existing curve loops using the provided transform.

    :param curve_loops: The curve loops to be transformed.
    :type curve_loops: list of Autodesk.Revit.DB.CurveLoop
    :param transform: The transform to be applied to the curve loops.
    :type transform: Autodesk.Revit.DB.Transform
    :param convert_net_list: Flag to indicate whether to convert the curve loops to a .net List of curves (true) or python list (false).
    :type convert_net_list: bool

    :return: A list of transformed curve loops.
    :rtype: .net List of Autodesk.Revit.DB.CurveLoop or python list of Autodesk.Revit.DB.CurveLoop
    """

    if convert_net_list:
        # convert the curve loops to a list of curves
        new_curve_loops = List[CurveLoop]()
        for curve_loop in curve_loops:
            # create a new curve loop
            new_curve_loop = CurveLoop.CreateViaTransform(curve_loop, transform)
            # add the new curve loop to the list
            new_curve_loops.Add(new_curve_loop)
        return new_curve_loops
    else:
        new_curve_loops = []
        for curve_loop in curve_loops:
            transformed_curve_loop = CurveLoop.CreateViaTransform(curve_loop,  transform)
            new_curve_loops.append(transformed_curve_loop)
        
        return new_curve_loops
    
    
def create_curve_loop_through_offset(curve_loop, offset_distance, offset_to_inside = True):
    """
    Creates a new curve loop by offsetting it a given distance.

    Some 

    :param curve_loop: The curve loop to be transformed.
    :type curve_loop: Autodesk.Revit.DB.CurveLoop
    :param offset_distance: The distance to offset the curve loop.
    :type offset_distance: float
    :param offset_normal: The normal vector to offset the curve loop.
    :type offset_normal: Autodesk.Revit.DB.XYZ

    :return: A transformed curve loops.
    :rtype: Autodesk.Revit.DB.CurveLoop
    """

    # Function to determine if a CurveLoop is an outer loop or an inner loop
    def is_outer_loop(curve_loop):
        # Calculate the area of the CurveLoop
        area = 0.0
        for curve in curve_loop:
            start = curve.GetEndPoint(0)
            end = curve.GetEndPoint(1)
            area += (start.X * end.Y - end.X * start.Y)
        return area > 0

    
    # Determine the offset direction for the outer loop
    if is_outer_loop(curve_loop):
        # If the loop is an outer loop, offset towards the inside
        if  offset_to_inside:
            offset_distance = -abs(offset_distance)
        else:
            offset_distance = abs(offset_distance)
    else:
        if offset_to_inside:
            offset_distance = abs(offset_distance)
        else:
            offset_distance = -abs(offset_distance)

    # Check if the curve loop is a circle
    curve_loop_is_circle = True
    # Check if the curve is arc and adjust the offset direction if necessary
    for curve in curve_loop:
        if not type(curve) == Arc:
            curve_loop_is_circle = curve_loop_is_circle and False
            break

    # If the curve loop is a circle, offset in the opposite direction
    if curve_loop_is_circle:
        offset_distance = -1 * offset_distance

    # convert to imperial feet
    offset_distance = convert_mm_to_imperial_feet(offset_distance)

    # create a new curve loop via offset
    new_curve_loop = CurveLoop.CreateViaOffset(curve_loop, offset_distance, XYZ.BasisZ)

    return new_curve_loop


def get_area_from_closed_curve_loop(doc, view, curve_loop, filled_region_type_id):
    """
    Calculates the area of a closed curve loop in a given view.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view in which the curve loop is located.
    :type view: Autodesk.Revit.DB.View
    :param curve_loop: The closed curve loop.
    :type curve_loop: Autodesk.Revit.DB.CurveLoop

    :return: The area of the closed curve loop.
    :rtype: float
    """

    # set up a status tracker
    return_value = Result()
    
    # default area value
    area = -1.0
    
    try:
        
        # check if the curve loop is closed
        if curve_loop.IsOpen():
            return_value.update_sep(False, "Curve loop is not closed.")
            return_value.result.append(area)
            return return_value
        
        curve_loop_net_list = List[CurveLoop]()
        curve_loop_net_list.Add(curve_loop)

        # set up an action creating a filled region using the loop
        def action():
            # set up a status tracker
            action_return_value = Result()
            try:
                # create the filled region using the loop
                action_return_value = create_filled_region_by_view(
                    doc=doc, 
                    view=view, 
                    curve_loops=curve_loop_net_list, 
                    filled_region_type=filled_region_type_id,
                    transaction_manager=None,
                )
                
            except Exception as e:
                action_return_value.update_sep(False, "Failed to get filled region area with error: {}".format(e))
            return action_return_value
    
        transaction = Transaction(doc, "Getting area of curve loop")
        return_value = in_transaction(transaction, action)

        if return_value.status == False:
            return return_value
        
        # get the actual filled region instance
        filled_region = return_value.result[0]
        
        # get the area of the filled region
        area = get_built_in_parameter_value(
            element=filled_region,
            built_in_parameter_def=BuiltInParameter.HOST_AREA_COMPUTED,
            parameter_value_getter=getter_double_as_double_converted_to_metric,
        )

        # store the area in the return value
        return_value.result=[area]

        # delete the filled region
        delete_result = delete_by_element_ids(
            doc=doc, 
            ids=[filled_region.Id],
            transaction_name="Deleting temp filled region",
            element_name="filled region",
        )

        if delete_result.status == False:
            return_value.update_sep(False, "Failed to delete filled region with error: {}".format(delete_result.message))
            return return_value
        
    except Exception as e:
        return_value.update_sep(False, "Failed to get filled region loop area with error: {}".format(e))
        
    return return_value


def convert_loop_to_curve_array(curve_loop):
    """
    Convert a curve loop to a CurveArray object.
    Curve loop can not contain nested loops.
    
    :param curve_loop: The curve loop to convert.
    :type curve_loop: Autodesk.Revit.DB.CurveLoop
    :return: CurveArray object
    """
    
    curve_array = CurveArray()
    if (isinstance(curve_loop, CurveLoop)):
        for curve in curve_loop:
            if (isinstance(curve, Curve)):
                curve_array.Append(curve)
                
    return curve_array


def convert_curve_loops_to_curve_arr_array(curve_loops):
    """
    Convert a list of curve loops to a CurveArrArray object.
    This is used to create a new extrusion in the family document.
    
    :param curve_loops: list of curve loops
    :return: CurveArrArray object
    """
    
    # create a new curve array array
    curve_arr_array = CurveArrArray()

    curve_array = CurveArray()
    for loop in curve_loops:
        
        if (isinstance(loop, CurveLoop)):
            # a loop of curves, convert to curve array
            c_ar = convert_loop_to_curve_array(loop)
            curve_arr_array.Append(c_ar)
        elif (isinstance(loop, Curve)):
            # just a curve, not a loop
            curve_array.Append(loop)
            
    # only append the curve array if it is not empty
    if (curve_array.Size > 0):
        curve_arr_array.Append(curve_array)

    return curve_arr_array


def get_curve_loop_centroid(curve_loop):
    """
    Finds an approximate centroid inside the given CurveLoop. The curve loop is assumed to be closed and planar.
    
    :param original_loop: The CurveLoop to find the centroid of.
    :type original_loop: Autodesk.Revit.DB.CurveLoop
    :return: The centroid point of the CurveLoop.
    :rtype: Autodesk.Revit.DB.XYZ
    """
    
    # check if the curve loop is closed
    if curve_loop.IsOpen():
        raise ValueError("Curve loop is not closed.")
        
    sum_point = XYZ.Zero
    count = 0

    for curve in curve_loop:
        sum_point += curve.GetEndPoint(0)
        sum_point += curve.GetEndPoint(1)
        count += 2

    centroid = sum_point / count
    return centroid  # Likely inside the original loop


def get_curve_loop_by_offset_towards_centroid( curve_loop, offset_distance):
    """
    Creates a new curve loop by offsetting the original curve loop towards its centroid.
    This is not a size change but a position change!
    
    :param curve_loop: The original CurveLoop to be offset.
    :type curve_loop: Autodesk.Revit.DB.CurveLoop
    :param offset_distance: The distance to offset the curve loop.
    :type offset_distance: float

    :return: A new CurveLoop that is offset from the original.
    :rtype: Autodesk.Revit.DB.CurveLoop
    """
    
    # get the centroid of the curve loop
    centroid = get_curve_loop_centroid(curve_loop)
    
    vector_to_centroid = None

    for curve in curve_loop:
        # check if the curve is a line
        if not isinstance(curve, Curve):
            raise ValueError("Curve loop contains non-linear curves.")
        
        # create a vector from the centroid to the first point of the curve loop
        vector_to_centroid = centroid - curve.GetEndPoint(0)

        break

    # normalize the vector and multiply by the offset distance
    offset_vector = vector_to_centroid.Normalize() * offset_distance
    
    # create a transform using the offset vector
    transform = Transform.CreateTranslation(offset_vector)
    
    # create a new curve loop by transforming the original curve loop
    new_curve_loop = create_curve_loops_through_transform([curve_loop], transform, convert_net_list=False)[0]
    
    return new_curve_loop



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import clr
import os
import sys

from duHast.Utilities.Objects.result import Result

from System.Collections.Generic import List

def load_net_dll_path(dlls_to_load):
    """
    Loads dlls from the libs folder of the duHast extension.
    Useful if a dll is not loaded through the startup script of a pyRevit extension.

    :param dlls_to_load: List of dlls to load.
    :type dlls_to_load: list
    :return: Result object with status and message.
    :rtype: Result
    """

    # set up a status tracker
    return_value = Result()

    try:

        # type checking
        if not isinstance(dlls_to_load, list):
            return_value.update_sep(
                False, "dlls_to_load must be a list, got {} instead.".format(type(dlls_to_load))
            )
            return return_value

        # the dll is located in the libs folder of the extension, which is one level up from the current file's directory
        current_directory = os.path.dirname(__file__)
        return_value.append_message("Current directory: {}".format(current_directory))
        parent_directory = os.path.dirname(current_directory)
        return_value.append_message("Parent directory: {}".format(parent_directory))

        # get all the lib paths from the sys.path
        lib_paths = [path for path in sys.path if path.endswith("lib")]

        if not lib_paths or len(lib_paths) == 0:
            return_value.update_sep(
                False, "No lib paths found in sys.path."
            )

            for p in sys.path:
                return_value.append_message("sys.path entry: {}".format(p))

            return return_value
        else:
            for p in lib_paths:
                duHast_path = os.path.join(p, "duHast")
                # check if path exists
                if os.path.exists(duHast_path):
                    return_value.append_message("Valid duHast path: {}".format(duHast_path))
                    # add the library path within the duHast folder to the sys.path
                    duHast_lib_path = os.path.join(duHast_path, "lib")
                    
                    # valid path check
                    if os.path.exists(duHast_lib_path):
                        return_value.append_message("Valid duHast//lib path: {}".format(duHast_lib_path))
                        
                        # load dlls
                        for dll in dlls_to_load:
                            # add the wrapper dll to the clr
                            dll_path = os.path.join(duHast_lib_path,dll)
                            
                            # valid path check
                            if os.path.exists( dll_path):
                                try:
                                    # add the dll to the clr
                                    clr.AddReferenceToFileAndPath(dll_path)
                                    return_value.append_message("Loaded dll: {}".format(dll_path))
                                except Exception as e:
                                    return_value.update_sep(
                                        False, "Failed to load dll: {}. Error: {}".format(dll_path, e)
                                    )
                            else:
                                return_value.update_sep(
                                    False, "DLL path does not exist: {}".format(dll_path)
                                )
                    else:
                        return_value.update_sep(
                            False, "Path to duHast//lib does not exist: {}".format(duHast_lib_path)
                        )
                        return return_value
                    break
                else:
                    return_value.update_sep(
                        False, "Path to duHast does not exist in : {}".format(p)
                    )
        return_value.append_message( "DLL path set successfully.")
    except Exception as e:
        # handle any exceptions that occur during the export process
        message = "An error occurred while processing export settings: {}".format(e)
        return_value.update_sep(
            False, message
        )


    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family shared parameter data class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Utilities import utility as util
from duHast.Revit.SharedParameters import shared_parameters as rSharedPara
from duHast.Revit.SharedParameters.Data.Objects.shared_parameter_data_storage import (
    FamilySharedParameterDataStorage,
)
from duHast.Revit.SharedParameters.Data.Objects.shared_parameter_storage_used_by import (
    FamilySharedParameterDataStorageUsedBy,
)


# import Autodesk
from Autodesk.Revit.DB import Element


# PARAMETER_GUID = "parameterGUID"
# PARAMETER_NAME = "parameterName"
# PARAMETER_ID = "parameterId"


class SharedParameterData(IFamData.IFamilyData):
    def __init__(self, root_path=None, root_category_path=None):
        """
        Initialize the shared parameter data object.

        :param root_path: Root path of the family.
        :type root_path: str
        :param root_category_path: Root category path of the family.
        :type root_category_path: str
        """

        super(SharedParameterData, self).__init__(
            root_path=root_path,
            root_category_path=root_category_path,
        )
        # super(CategoryData, self).__init__(rootPath, dataType)

    def process(self, doc):
        """
        Process the document and find all shared parameters in the family.
        Update the data storage with the shared parameter data.

        :param doc: Revit family document
        :type doc: Autodesk.Revit.DB.Document
        :return: None
        """

        # make sure to get a value for the file path which is not empty if the document has not been saved
        saved_file_name = "-"
        if doc.PathName != "":
            saved_file_name = doc.PathName

        collector = rSharedPara.get_all_shared_parameters(doc)
        for para in collector:
            # just in case parameter name is not unicode
            parameter_name = "unknown"
            try:
                parameter_name = util.encode_ascii(Element.Name.GetValue(para))
            except Exception as ex:
                parameter_name = "Exception: {}".format(ex)
            # check if used:
            use_counter = 0
            used_by_list = []
            if rSharedPara.is_shared_parameter_definition_used(doc, para):
                # this will be 1 at most
                use_counter = use_counter + 1
                # build used by data as required to be the same as post process update
                used_by_data = FamilySharedParameterDataStorageUsedBy(
                    family_name=self.root_path,
                    element_id=para.Id.IntegerValue,
                )
                used_by_list.append(used_by_data)

            # build data
            storage = FamilySharedParameterDataStorage(
                root_name_path=self.root_path,
                root_category_path=self.root_category_path,
                family_name=self._strip_file_extension(doc.Title),
                family_file_path=saved_file_name,
                parameter_guid=para.GuidValue.ToString(),
                parameter_name=parameter_name,
                parameter_id=para.Id.IntegerValue,
                use_counter=use_counter,
                used_by=used_by_list,
            )
            self.add_data(storage_instance=storage)

        # check if any shared parameter was found
        if len(self.data) == 0:
            # add message no shared parameter found
            storage = FamilySharedParameterDataStorage(
                root_name_path=self.root_path,
                root_category_path=self.root_category_path,
                family_name=self._strip_file_extension(doc.Title),
                family_file_path=saved_file_name,
                parameter_guid="",
                parameter_name="No shared parameter present in family.",
                parameter_id=-1,
                use_counter=0,
                used_by=[],
            )
            self.add_data(storage_instance=storage)

    def get_data(self):
        """
        Get the shared parameter data.

        :return: List of shared parameter data storage instances.
        """

        return self.data

    def add_data(self, storage_instance):
        """
        Add shared parameter data to the data storage.

        :param storage_instance: Shared parameter data storage instance.
        :type storage_instance: FamilySharedParameterDataStorage
        :return: None
        """

        if isinstance(storage_instance, FamilySharedParameterDataStorage):
            self.data.append(storage_instance)
        else:
            raise ValueError(
                "storage instance must be an instance of FamilySharedParameterDataStorage"
            )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family category purge unused utilities.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This will delete all subcategories which are user created ( id greater then 0) and are not used by any element in the family or nested families.

- requires a revit category processor object

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# class used for stats reporting
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common import delete as rDel
from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Revit.Categories.Data.Objects import category_data as rCatData

from Autodesk.Revit.DB import ElementId


def purge_unused_sub_categories(doc, processor):
    """
    This function deletes all user-created subcategories (with IDs greater than 0) that are not used by any element in the family or nested families.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param processor: A CategoryProcessor object containing all subcategory information of the family document and any nested families.
    :type processor: :class:`.CategoryProcessor`

    :return: Result class instance.
        - `result.status` (bool): True if all unused subcategories were deleted successfully or none needed to be deleted, otherwise False.
        - `result.message` (str): Property updated in the format: "Found unused subcategory: family category Name : subcategory Name [subcategory Id]"

    On exception:
        - `status` (bool): False.
        - `message` (str): Contains the exception message.

    :rtype: :class:`.Result`
    """

    # from processor instance get all root category entries where usage counter == 0 and subCategoryId > 0 (pointing to a custom sub category and not a built in one).
    # delete those subcategories by id

    return_value = res.Result()

    ids_to_delete = []
    # get categories found in root processor data only
    root_fam_data = processor._find_root_family_data()
    # get all root category entries where usage counter == 0 and subCategoryId > 0 (pointing to a custom sub category and not a built in one).
    for root_fam in root_fam_data:
        if (
            root_fam[IFamData.USAGE_COUNTER] == 0
            and root_fam[rCatData.SUB_CATEGORY_ID] > 0
        ):
            return_value.append_message(
                "Found unused sub category: {} : {} [{}]".format(
                    root_fam[rCatData.CATEGORY_NAME],
                    root_fam[rCatData.SUB_CATEGORY_NAME],
                    (root_fam[rCatData.SUB_CATEGORY_ID]),
                )
            )
            ids_to_delete.append(ElementId(root_fam[rCatData.SUB_CATEGORY_ID]))
    # delete any subcategories found
    if len(ids_to_delete) > 0:
        result_delete = rDel.delete_by_element_ids(
            doc, ids_to_delete, "Deleting unused sub categories.", "Subcategories"
        )
        return_value.update(result_delete)
    else:
        return_value.update_sep(
            True, "No unused categories found. Nothing was deleted."
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit floors helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

from duHast.Revit.Common import common as com
from duHast.Revit.Family import family_utils as rFam

# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInCategory,
    ElementCategoryFilter,
    FilteredElementCollector,
    FamilyInstance,
    Floor,
)

from duHast.Revit.Floors.Utility import floors_filter as rFloorFilter

#: Built in family name for standard floor
FLOOR_FAMILY_NAME = "Floor"
#: Built in family name for a foundation slab
FOUNDATION_SLAB_FAMILY_NAME = "Foundation Slab"

#: List of all Built in floor family names
BUILTIN_FLOOR_TYPE_FAMILY_NAMES = [FLOOR_FAMILY_NAME, FOUNDATION_SLAB_FAMILY_NAME]

# --------------------------------------------- utility functions ------------------


def get_all_floor_types_by_category(doc):
    """
    Gets a filtered element collector of all ceiling types in the model:

    - Compound Ceiling
    - In place families or loaded families
    - Basic Ceiling

    Filters by category.
    It will therefore not return any roof soffit types ..

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing ceiling types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rFloorFilter._get_all_ceiling_types_by_category(doc)
    return collector


def get_floor_types_by_class(doc):
    """
    Gets a filtered element collector of all ceiling types in the model:

    - Roof Soffit
    - Compound Ceiling
    - Basic Ceiling

    Filters by class.
    It will therefore not return any in place family types.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing ceiling types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rFloorFilter._get_floor_types_by_class(doc)
    return collector


# -------------------------------- none in place Floor types -------------------------------------------------------


def get_all_floor_instances_in_model_by_category(doc):
    """
    Gets all floor elements placed in model...ignores in foundation slabs.

    Filters by builtin category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing floor instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Floors)
        .WhereElementIsNotElementType()
    )


def get_all_floor_instances_in_model_by_class(doc):
    """
    Gets all floor elements placed in model...ignores in place families of category floor.

    Filters by class.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing floor instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return FilteredElementCollector(doc).OfClass(Floor).WhereElementIsNotElementType()


def get_all_floor_type_ids_in_model_by_category(doc):
    """
    Returns all Floor element types available in model.

    Filters by builtin category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing floor types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    ids = []
    col_cat = get_all_floor_types_by_category(doc)
    ids = com.get_ids_from_element_collector(col_cat)
    return ids


def get_all_floor_type_ids_in_model_by_class(doc):
    """
    Returns all Floor element types available in model.

    Filters by class.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing floor types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    ids = []
    col_class = get_floor_types_by_class(doc)
    ids = com.get_ids_from_element_collector(col_class)
    return ids


# -------------------------------- In place Floor types -------------------------------------------------------


def get_in_place_floor_family_instances(doc):
    """
    Gets all instances of in place families of category floor.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing floor family instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    filter = ElementCategoryFilter(BuiltInCategory.OST_Floors)
    return FilteredElementCollector(doc).OfClass(FamilyInstance).WherePasses(filter)


def get_all_in_place_floor_type_ids_in_model(doc):
    """
    Gets type ids off all available in place families symbols (types) of category floor.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing in place floor symbols (types).
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = rFam.get_all_in_place_type_ids_in_model_of_category(
        doc, BuiltInCategory.OST_Floors
    )
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit room separation lines reports functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.Common.design_set_options import get_design_set_option_info
from duHast.Revit.Common.Geometry.curve import get_curve_level
from duHast.Revit.Common.Objects.design_set_property_names import DesignSetPropertyNames
from duHast.Revit.Rooms.room_lines import get_all_room_separation_lines_ids_with_warnings, get_all_room_separation_lines_ids_without_warnings



def _sort_room_lines_by_design_option(doc, room_line_ids):
    """
    Sorts room lines into a dictionary where key is the combination of the design set name and
    design option name and value is a list of room separation lines.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param room_line_ids: list of room separation line ids
    :type room_line_ids: [Autodesk.Revit.DB.ElementId]

    :return: Dictionary containing room separation lines by design set / option name
    :rtype: {str:[Autodesk.Revit.DB.ModelCurve]}
    """

    room_lines_by_design_option = {}

    # loop over ids, get the line and its design set / option data
    for id in room_line_ids:
        element = doc.GetElement(id)
        data_dic = get_design_set_option_info(doc=doc, element=element)
        key = DesignSetPropertyNames.combine_set_and_option_name(set_name=data_dic[DesignSetPropertyNames.DESIGN_SET_NAME], option_name=data_dic[DesignSetPropertyNames.DESIGN_OPTION_NAME])
        if key in room_lines_by_design_option:
            room_lines_by_design_option[key].append(element)
        else:
            room_lines_by_design_option[key] = [element]

    return room_lines_by_design_option


def _sort_room_lines_by_level(doc, room_lines_by_design_option):
    """
    Sorts the lines by level

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param room_lines_by_design_option: A dictioanry where key is the design set / option name and value is a list of room separation lines
    :type room_lines_by_design_option: {str:[Autodesk.Revit.DB.ModelCurve]}

    :return: A dictionary where key is a dictioanry where key is the design set / option name and value is another dictionary where key is the level name and value is a list of room separation lines
    :rtype: {str:{str:[Autodesk.Revit.DB.ModelCurve]}}
    """
    room_lines_by_design_option_and_level = {}

    # loop over dictionary and get the level information per line
    for design_set_option, lines in room_lines_by_design_option.items():
        curves_by_level_name = {}
        for line in lines:
            level = get_curve_level(doc=doc, curve=line)
            level_name = "Line is not associated to any level"
            if level:
                level_name = level.Name
            if level_name in curves_by_level_name:
                curves_by_level_name[level_name].append(line)
            else:
                curves_by_level_name[level_name] = [line]
        room_lines_by_design_option_and_level[design_set_option] = curves_by_level_name

    return room_lines_by_design_option_and_level



def _sort_room_sep_lines_by_design_option_and_level(doc, rooms_sep_line_ids):
    """
    Sort room lines by design option and by level

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param rooms_sep_line_ids: List of room separation line ids
    :type rooms_sep_line_ids: [Autodesk.Revit.ElementId]

    :return: A dictionary where key is the design option name and value is a nested dictionary where key is the level name and value is a list of room lines.
    :rtype: :rtype: {str: {str:[Autodesk.Revit.DB.ModelCurve]}}
    """

    # sort offending elements by design option
    room_lines_by_design_option = _sort_room_lines_by_design_option(
        doc=doc, room_line_ids=rooms_sep_line_ids
    )

    # sort offending elements by level
    room_line_instances_by_design_option_and_level = _sort_room_lines_by_level(
        doc=doc, room_lines_by_design_option=room_lines_by_design_option
    )

    return room_line_instances_by_design_option_and_level


def room_lines_with_warnings_by_design_option_and_level(doc):
    """
    Reports all room separation line with warnings by design option and level they belong too

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: dictionary where key is the design option name and value is a nested dictionary where key is the level name and value is a list of room lines.
    :rtype: {str: {str:[Autodesk.Revit.DB.ModelCurve]}}
    """

    room_line_instances_by_design_option_and_level = {}

    # get all ids which have warnings
    all_line_ids_with_warnings = get_all_room_separation_lines_ids_with_warnings(doc=doc)

    # check if any warning is present
    if len(all_line_ids_with_warnings) == 0:
        return room_line_instances_by_design_option_and_level

    # sort by design option and level
    room_line_instances_by_design_option_and_level = _sort_room_sep_lines_by_design_option_and_level (doc=doc, rooms_sep_line_ids= all_line_ids_with_warnings)
   

    return room_line_instances_by_design_option_and_level


def room_lines_without_warnings_by_design_option_and_level(doc):

    """
    Reports all room separation line without warnings by design option and level they belong too

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary where key is the design option name and value is a nested dictionary where key is the level name and value is a list of room lines.
    :rtype: {str: {str:[Autodesk.Revit.DB.ModelCurve]}}
    """

    room_line_instances_by_design_option_and_level = {}


    # get all room separation lines without warnings
    all_line_ids_without_warnings = get_all_room_separation_lines_ids_without_warnings(doc=doc)
    
    # check if any warning is present
    if len(all_line_ids_without_warnings) == 0:
        return room_line_instances_by_design_option_and_level

    # sort by design option and level
    room_line_instances_by_design_option_and_level = _sort_room_sep_lines_by_design_option_and_level (doc=doc, rooms_sep_line_ids= all_line_ids_without_warnings)
   

    return room_line_instances_by_design_option_and_level
  